/*******************************************************************************/
//
//		Midi Controlled Light Bulb Display
//		Created by: Colin Mann
//		September, 2010
//
//
/*******************************************************************************/

#include <avr/io.h>
#include <avr/interrupt.h>
#include <stdlib.h>
#include "hardware.h"
#include <util/delay.h>
#include "messages.h"
#include "uart.h"

void usb_analyze(unsigned char);
void midi_analyze(unsigned char);
void sysexAnalyze(unsigned char);
void setRelay(int, int);
void lightRun(void);

volatile int time=0;

//Midi Data
unsigned char current_status_byte = 0;
unsigned char message_position = 0;
unsigned char data_byte = 0;
unsigned char current_control_byte = 0;			//Used only for sustain pedal commands, possibly somethine else later.
unsigned char note_hold = 0;					//Sustain Pedal off.
unsigned char sustain_notes[12];				//Notes being held by sustain pedal.
unsigned char sysexVerified = 0;
unsigned char current_sysex_byte = 0;
volatile unsigned char mode = COMBO_MODE;
unsigned char home_position = 0;
volatile unsigned char prev_mode = COMBO_MODE;

//Switch Data
int switch_debounce_count = 0;
unsigned char switch_state = 0;

int main(void)
{
	hardwareInit();
	
	_delay_ms(1000);
	uart_init( UART_BAUD_SELECT(USB_BAUD_RATE,F_CPU) ); 
	uart1_init( UART_BAUD_SELECT(MIDI_BAUD_RATE,F_CPU) ); 

	sei();

	lightRun();

	unsigned int midi_byte;
	unsigned int usb_byte;

	while(1)
	{			
		
		//Check for button press, for setting the home position.
		if( bit_is_clear(SW_PIN,SWITCH) && !(switch_state) )
		{
			if( (switch_debounce_count==DEBOUNCE_COUNT) )
			{
				if(mode!=SETHOME)
				{
					prev_mode = mode;
					mode = SETHOME;
					switch_debounce_count = 0;
					switch_state=1;
					uart_putc(DEBUG);
					uart_putc(switch_debounce_count);
				}
			}
			else
				switch_debounce_count++;
			
		}
		else
			switch_state=0;


		usb_byte = uart_getc();
		if(usb_byte & UART_NO_DATA);
		else
		{
			if ( usb_byte & UART_FRAME_ERROR )
			{
				uart_putc(DEBUG);
				uart_putc(USB_FRAME_ERROR);
			}
			else if ( usb_byte & UART_OVERRUN_ERROR )
			{
				uart_putc(DEBUG);
				uart_putc(USB_OVERRUN_ERROR);
			}
			else if ( usb_byte & UART_BUFFER_OVERFLOW )
			{
				uart_putc(DEBUG);
				uart_putc(USB_BUFFER_OVERFLOW);
			}
			else
				usb_analyze((unsigned char)usb_byte);
		}

		midi_byte = uart1_getc();
		if(midi_byte & UART_NO_DATA);
		else
		{
			if ( midi_byte & UART_FRAME_ERROR )
			{
				uart_putc(DEBUG);
				uart_putc(MIDI_FRAME_ERROR);
			}
			else if ( midi_byte & UART_OVERRUN_ERROR )
			{
				uart_putc(DEBUG);
				uart_putc(MIDI_OVERRUN_ERROR);
			}
			else if ( midi_byte & UART_BUFFER_OVERFLOW )
			{
				uart_putc(DEBUG);
				uart_putc(MIDI_BUFFER_OVERFLOW);
			}
			else
				midi_analyze((unsigned char)midi_byte);
		}



		//Led Indicator Settings - Based on current mode
		if(mode==LOCAL_MODE)				//Green
		{
			LED_PORT |= (1<<LED_RED);		
			LED_PORT &= ~(1<<LED_GRN);
			LED_PORT |= (1<<LED_BLU);
		}
		else if(mode == EXT_MODE)			//Red
		{
			LED_PORT &= ~(1<<LED_RED);
			LED_PORT |= (1<<LED_GRN);
			LED_PORT |= (1<<LED_BLU);
		}
		else if(mode == COMBO_MODE)			//Orange
		{
			LED_PORT &= ~(1<<LED_RED);
			LED_PORT &= ~(1<<LED_GRN);
			LED_PORT |= (1<<LED_BLU);
		}
		else if(mode == SETHOME)			//Blue
		{
			LED_PORT |= (1<<LED_RED);
			LED_PORT |= (1<<LED_GRN);
			LED_PORT &= ~(1<<LED_BLU);
		}
	}

	return 0;
}

void usb_analyze(unsigned char c)
{
	unsigned char type = (c&0xF0);
	if(type==BULBOFF)
		setRelay(((12+(c&0x0F))-home_position)%12,0);
	else if(type==BULBON)
	{
		if(mode==SETHOME)
		{
			home_position = c&0x0F;
			mode = prev_mode;
		}
		else
			setRelay(((12+(c&0x0F))-home_position)%12,1);
	}
	else if(c==PEDALON)
		note_hold = 1;
	else if(c==PEDALOFF)
	{
		note_hold = 0;
		for(c=0;c<12;c++)
		{
			if(sustain_notes[c]==1)
			{
				setRelay(c,0);
				sustain_notes[c] = 0;
			}
		}
	}
	else if(type==CLEAR)
	{
		RELAYPORT0 = (0xFF)*(c&0x0F);
		RELAYPORT1 = (0xFF)*(c&0x0F);
	}
	else if(type==HOMESET)
		home_position = (c&0x0F);
	else if(c==RUNINIT)
		lightRun();
	else if(type==SETMODE)
	{
		prev_mode = mode;
		mode = (c&0x0F)%4;
	}
}

void midi_analyze(unsigned char c)
{
	//Check for status byte
	if(MIDI_STATUS & c)
	{
		//Ignore midi clock and active sense data.
		if(c > MIDI_SYSEX)
			return;
		else
			current_status_byte = c & 0xF0;		//Eliminate channel data, also, maintains the sysex message
	}
	else
	{
		switch(current_status_byte)
		{
			case MIDI_NOTE_OFF:
				if(message_position)
				{
					setRelay(data_byte,0);
					message_position = 0;
				}
				else
				{
					message_position=1;
					data_byte = ((12+(c%12))-home_position)%12;
				}
				break;
			case MIDI_NOTE_ON:
				if(message_position)
				{
					if(mode==SETHOME)
					{
						home_position = data_byte;
						mode = prev_mode;
					}
					else
						setRelay(data_byte,c);
					message_position = 0;
				}
				else
				{
					message_position=1;
					data_byte = ((12+(c%12))-home_position)%12;
				}
				break;
			case MIDI_POLY_PRES:
				if(message_position)
					message_position = 0;
				else
					message_position = 1;
				break;
			case MIDI_CC:
				if(message_position)
				{
					if(current_control_byte == MIDI_SUSTAIN)
					{
						if(c<63)
						{
							note_hold = 0;
							for(c=0;c<12;c++)
							{
								if(sustain_notes[c]==1)
								{
									setRelay(c,0);
									sustain_notes[c] = 0;
								}
							}
						}
						else
							note_hold = 1;
					}
					message_position = 0;
				}
				else
				{
					message_position = 1;
					current_control_byte = c;
				}
				break;	
			case MIDI_PC:
				message_position = 0;
				break;
			case MIDI_CHAN_PRES:
				message_position = 0;
				break;
			case MIDI_PITCH:
				if(message_position)
					message_position = 0;
				else
					message_position = 1;
				break;
			case MIDI_SYSEX:
				if(message_position)
				{
					if(c == MIDI_SYSEX_END)
						message_position = 0;
					else if(sysexVerified)
					{
						sysexAnalyze(c);
					}
				}
				else
				{
					message_position = 1;
					if(c == SYS_ID)
						sysexVerified=1;
					else
						sysexVerified=0;
				}
				break;
		}
	}
}

void sysexAnalyze(unsigned char message)
{
	int c=0;
	if(current_sysex_byte==SETHOME)
	{
		home_position = message%12;		//Set Home Position
		current_sysex_byte = 0;
	}
	else
	{
		current_sysex_byte = message;
		mode = message;
		for(c=0;c<12;c++)
		{
			setRelay(c,0);
			sustain_notes[c] = 0;
		}
	}

}



void setRelay(int channel, int state)
{
	if(state)
	{
		if(channel<6)
			RELAYPORT0 |= (1<<channel);
		else
			RELAYPORT1 |= (1<<(channel-6));
	}
	else
	{
		if(note_hold)
			sustain_notes[channel] = 1;
		else
		{
			if(channel<6)
				RELAYPORT0 &= ~(1<<channel);
			else
				RELAYPORT1 &= ~(1<<(channel-6));
		}
	}
}

void lightRun()
{
	int i=0;
	for(i=0;i<12;i++)			//Clear Bulbs
		setRelay(i,0);
	for(i=0;i<12;i++)			//Do Run
	{
		setRelay(i,1);
		_delay_ms(100);
	}
	for(i=0;i<12;i++)			//Undo run, leaves bulbs in off state.
	{
		setRelay(i,0);
		_delay_ms(100);
	}
}

